/*
 * Le Lann, Chang and Roberts leader election algorithm for a ring of
 * processes
 *
 * This solution contains a 'random' initialization procedure:
 * - computes a random base value j, and initialize a vector of identifiers
 *   ids = [j, ..., j+N-1]; since j is unknown to the nodes they can't use
 *   the knowledge of N to individually compute who is (and who isn't) the leader.
 * - performs (at most) N swap operations, by randomly selecting one of the two array indexes,
 *   so that the values in the ids vector are not (necessarily) monotonically increasing anymore.
 *   This prevents the nodes from using _pid and N to individually compute who is (and who isn't)
 *   the leader.
 *
 * Although this code 'works' and it does not contain 'infinite loops', if you
 * attempt to run its associated verifier then you can see that you can't prove
 * its correctness (unless you own some really powerful machine):
 *
 * ~$ spin -a leader_lcr_random_01.pml
 * ~$ gcc pan.c -o random01.x
 * ~$ ./random01.x
 *        Depth=     380 States=    1e+06 Transitions= 1.58e+06 Memory=   311.347	t=     2.17 R=   5e+05
 *        ...
 *        Depth=     380 States=  7.6e+07 Transitions=  1.2e+08 Memory= 14508.308	t=      190 R=   4e+05
 *    process killed: too much memory
 *
 * On my notebook, before the process is automatically killed by the OS, the verifier
 * generates up to 7.6*10^7 states and 1.2*10^8 transitions, filling up 14.5 GB of memory.
 *
 * WARNING: repeat this experiment at your own risk! if you have a swap partition, your OS
 *     might decide to start swapping memory instead of killing the process. This might
 *     'freeze' your machine for several minutes [or hours], even if you type a last-minute CTRL+C.
 *
 * This code example highligths the limitation of this model checker due to the particular
 * encoding representation adopted for the input problem.
 *
 * The state space explosion is caused by two parts of this code, the memory-swap
 *                byte tmp = ids[i];
 *                ids[i] = ids[j];
 *                ids[j] = tmp;
 * and the loop setting j to a random value:
 *        j = 0;
 *        do
 *            :: j < J_BOUND ->
 *                j++
 *            :: true ->
 *                break;
 *        od;
 *
 * The loop setting j is 'heavy' on its own (as you can see by playing with leader_lcr_random_02.pml),
 * but the 'real' problem arises by the representation of swap operations with the large pool of
 * values that j can have. Remember that a verifier checks the entire search space and every possible
 * execution!
 *
 * If you try to decrease the value of J_BOUND, you can notice that it effectively decreases
 * the search space size. After a certain (platform-dependent) threshold, the verifier will
 * be able to terminate successfully (confirming that the code is 'good'). You can also set
 * J_BOUND to a fixed value, and disable the 'swap code' to achieve a dramatic gain in performance. :-)
 *
 */

#define N 5
#define J_BOUND (255 - N - 1) // Try with: 100, 50, 10
#define BUFSIZE 10

mtype = { candidate, leader };
chan c[N] = [BUFSIZE] of { mtype, byte };

byte num_leaders = 0;

proctype node(chan prev, next; byte my_id)
{
    byte in_id;
    byte leader_id;

    // init: pass the node id to its neighbour
    next!candidate(my_id);

    do
        :: prev?candidate(in_id) ->
            if
                :: in_id > my_id ->
                    next!candidate(in_id);
                :: in_id < my_id ->
                    skip;
                :: else ->
                    printf("I am the leader! (pid: %d, id: %d)\n", _pid, my_id);
                    leader_id = my_id;
                    num_leaders++;
                    next!leader(my_id);
            fi
        :: prev?leader(leader_id) ->
            if
                :: leader_id > my_id ->
                    next!leader(leader_id);
                    break
                :: else ->
                    assert(my_id == leader_id);
                    break
            fi
    od;
    assert(num_leaders == 1); // NOTE: if more than one node
                              // declares itself as leader, then
                              // this assertion will fail for
                              // at least one node!
}

init {
    byte ids[N];
    byte i, j;

    atomic {
        // pick up random j
        j = 0;
        do
            :: j < J_BOUND ->
                j++;
            :: true ->
                break;
        od;
        // create unique ids
        do
            :: i < N ->
               ids[i] = i + j;
               i++;
            :: else ->
               break;
        od;
        // swap each i with a random j
        i = 0;
        do
            :: i < N ->
                // randomly select j
                j = 1;
                do
                    :: j < N ->
                        j++;
                    :: true ->
                        break;
                od;
                j=(j%N);
                // swap elements
                byte tmp = ids[i];
                ids[i] = ids[j];
                ids[j] = tmp;
                // handle next memory location
                i++;
            :: else ->
               break;
        od;
        i = 0;
        do
            :: i < N ->
                printf("Starting process with id %d.\n", ids[i]);
                run node(c[i], c[(i+1)%N], ids[i]);
                i++;
            :: else ->
                break;
        od
    }
}
